/*
 * Copyright 2009-2013 by The Regents of the University of California
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * you may obtain a copy of the License from
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package edu.uci.ics.asterix.transaction.management.test;

import java.io.IOException;
import java.util.Random;

import edu.uci.ics.asterix.common.config.AsterixPropertiesAccessor;
import edu.uci.ics.asterix.common.config.AsterixTransactionProperties;
import edu.uci.ics.asterix.common.exceptions.ACIDException;
import edu.uci.ics.asterix.common.exceptions.AsterixException;
import edu.uci.ics.asterix.common.transactions.DatasetId;
import edu.uci.ics.asterix.common.transactions.ILockManager;
import edu.uci.ics.asterix.common.transactions.ILogManager;
import edu.uci.ics.asterix.common.transactions.ILogger;
import edu.uci.ics.asterix.common.transactions.IRecoveryManager;
import edu.uci.ics.asterix.common.transactions.IResourceManager;
import edu.uci.ics.asterix.common.transactions.IResourceManager.ResourceType;
import edu.uci.ics.asterix.common.transactions.ITransactionContext;
import edu.uci.ics.asterix.common.transactions.ITransactionManager;
import edu.uci.ics.asterix.common.transactions.JobId;
import edu.uci.ics.asterix.common.transactions.LogUtil;
import edu.uci.ics.asterix.common.transactions.LogicalLogLocator;
import edu.uci.ics.asterix.transaction.management.logging.IResource;
import edu.uci.ics.asterix.transaction.management.service.logging.LogType;
import edu.uci.ics.asterix.transaction.management.service.transaction.JobIdFactory;
import edu.uci.ics.asterix.transaction.management.service.transaction.TransactionSubsystem;

public class TransactionSimulator {

    private ITransactionManager transactionManager;
    private ILogManager logManager;
    private ILockManager lockManager;
    private IRecoveryManager recoveryManager;
    private IResourceManager resourceMgr;
    private ILogger logger;
    private IResource resource;
    private LogicalLogLocator memLSN;
    private TransactionSubsystem transactionProvider;

    public TransactionSimulator(IResource resource, IResourceManager resourceMgr) throws ACIDException, AsterixException {
        String id = "nc1";
        transactionProvider = new TransactionSubsystem(id, null, new AsterixTransactionProperties(
                new AsterixPropertiesAccessor()));
        transactionManager = transactionProvider.getTransactionManager();
        logManager = transactionProvider.getLogManager();
        lockManager = transactionProvider.getLockManager();
        recoveryManager = transactionProvider.getRecoveryManager();
        transactionProvider.getTransactionalResourceRepository().registerTransactionalResourceManager(
                resourceMgr.getResourceManagerId(), resourceMgr);
        this.resourceMgr = resourceMgr;
        this.logger = resource.getLogger();
        this.resource = resource;
        memLSN = LogUtil.getDummyLogicalLogLocator(transactionProvider.getLogManager());
    }

    public ITransactionContext beginTransaction() throws ACIDException {
        JobId jobId = JobIdFactory.generateJobId();
        return transactionManager.beginTransaction(jobId);
    }

    public void executeTransactionOperation(ITransactionContext txnContext, FileResource.CounterOperation operation)
            throws ACIDException {
        // lockManager.lock(txnContext, resourceId, 0);
        ILogManager logManager = transactionProvider.getLogManager();
        int currentValue = ((FileResource) resource).getMemoryCounter();
        int finalValue;
        switch (operation) {
            case INCREMENT:
                finalValue = currentValue + 1;
                int logRecordLength = ((FileLogger) logger).generateLogRecordContent(currentValue, finalValue);
                logManager.log(LogType.UPDATE, txnContext, 1, -1, 1, ResourceType.LSM_BTREE, 0, null, logger, memLSN);
                ((FileResource) resource).increment();
                break;
            case DECREMENT:
                finalValue = currentValue - 1;
                logRecordLength = ((FileLogger) logger).generateLogRecordContent(currentValue, finalValue);
                logManager.log(LogType.UPDATE, txnContext, 1, -1, 1, ResourceType.LSM_BTREE, 0, null, logger, memLSN);
                ((FileResource) resource).decrement();
                break;
        }

    }

    public void commitTransaction(ITransactionContext context) throws ACIDException {
        transactionManager.commitTransaction(context, new DatasetId(-1), -1);
    }

    public void recover() throws ACIDException, IOException {
        recoveryManager.startRecovery(true);
        ((FileResource) resource).sync();
    }

    /**
     * @param args
     */
    public static void main(String[] args) throws IOException, ACIDException, AsterixException {
        String fileDir = "testdata";
        String fileName = "counterFile";
        IResource resource = new FileResource(fileDir, fileName);
        FileResourceManager resourceMgr = new FileResourceManager();
        resourceMgr.registerTransactionalResource(resource);
        int existingValue = ((FileResource) resource).getDiskCounter();

        TransactionSimulator txnSimulator = new TransactionSimulator(((FileResource) resource), resourceMgr);
        int numTransactions = 2;
        Schedule schedule = new Schedule(numTransactions);

        for (int i = 0; i < numTransactions; i++) {
            ITransactionContext context = txnSimulator.beginTransaction();
            txnSimulator.executeTransactionOperation(context, schedule.getOperations()[i]);
            if (schedule.getWillCommit()[i]) {
                txnSimulator.commitTransaction(context);
            }
        }

        int finalExpectedValue = existingValue + schedule.getDeltaChange();
        txnSimulator.recover();
        boolean isCorrect = ((FileResource) resource).checkIfValueInSync(finalExpectedValue);
        System.out.println(" Did recovery happen correctly " + isCorrect);
    }

}

class ResourceMgrIds {

    public static final byte FileResourceMgrId = 1;

}

class Schedule {

    private int numCommittedIncrements;
    private int numCommittedDecrements;

    private FileResource.CounterOperation[] operations;
    private Boolean[] willCommit;

    public Boolean[] getWillCommit() {
        return willCommit;
    }

    private Random random = new Random();

    public int getDeltaChange() {
        return numCommittedIncrements - numCommittedDecrements;
    }

    public Schedule(int numTransactions) {
        operations = new FileResource.CounterOperation[numTransactions];
        willCommit = new Boolean[numTransactions];
        for (int i = 0; i < numTransactions; i++) {
            willCommit[i] = random.nextBoolean();
            int nextOp = random.nextInt(2);
            FileResource.CounterOperation op = nextOp == 0 ? FileResource.CounterOperation.INCREMENT
                    : FileResource.CounterOperation.DECREMENT;
            operations[i] = op;
            if (willCommit[i]) {
                if (op.equals(FileResource.CounterOperation.INCREMENT)) {
                    numCommittedIncrements++;
                } else {
                    numCommittedDecrements++;
                }
            }
        }

    }

    public String toString() {
        StringBuilder builder = new StringBuilder();
        for (int i = 0; i < operations.length; i++) {
            builder.append(" operation " + operations[i]);
            if (willCommit[i]) {
                builder.append(" commit ");
            } else {
                builder.append(" abort ");
            }
        }

        builder.append(" number of committed increments " + numCommittedIncrements);
        builder.append(" number of committed decrements " + numCommittedDecrements);
        return new String(builder);
    }

    public FileResource.CounterOperation[] getOperations() {
        return operations;
    }

}
